C     SUBROUTINE IQLVMS
C     =================
C
C    GPLed 2002 by Gunter Winkler (guwi17@gmx.de)
C 
C    This program is free software; you can redistribute it and/or modify
C    it under the terms of the GNU General Public License as published by
C    the Free Software Foundation; either version 2 of the License, or
C    (at your option) any later version.
C 
C    This program is distributed in the hope that it will be useful,
C    but WITHOUT ANY WARRANTY; without even the implied warranty of
C    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C    GNU General Public License for more details.
C 
C    You should have received a copy of the GNU General Public License
C    along with this program; if not, write to the Free Software
C    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
C 
C    important info: This routine was taken from a numeric repository
C    of the Chemnitz University (http://www.tu-chemnitz.de) Many people
C    including Prof. Arndt Meyer and Dr. Matthias Pester developed this
C    code. They were inspired by similar EISPACK routines.
C
C ******************************************************************
C
C     PROGRAMMCHARAKTERISTIK
C     ----------------------
C
C IQLVMS BERECHNET MIT DEM IMPLIZITEN QL-ALGORITHMUS DIE
C EIGENWERTE UND EIGENVEKTOREN
C
C     BEDEUTUNG DER PARAMETER:
C
C       NM      (I) :   VEREINBARTE DIMENSION VON Z
C       N       (I) :   DIMENSION
C       D       (I) :   DIAGONALE DER TRIDIAGONALMATRIX
C               (O) :   EIGENWERTE
C       E       (I) :   NEBENDIAGONALE DER TRIDIAGONALMATRIX
C       Z       (I) :   EINIGE ZEILEN DER
C                       TRANSFORMATIONSMATRIX AUF TRIDIAGONALGESTALT
C               (O) :   EIGENVEKTORKOMPONENTEN DER URSPRUNGSMATRIX
C       IEV     (I) :   =0: KEINE EV-BERECHNUNG
C                       =K: JE K EV-KOMPONENTEN ALLER EIGENVEKTOREN
C       IER     (O) :   FEHLERANZEIGER;
C                       NUMMER DES EIGENWERTES, BEI DEM DIE MAXIMALE
C                       ITERATIONSZAHL VON 30 ERREICHT WURDE;
C                       (IER-1) EIGENWERTE SIND KORREKT
C
C **********************************************************************
C
      SUBROUTINE IQLVMS(NM,N,D,E,Z,IEV,IER)
      IMPLICIT REAL*8 (A-H,O-Z)
      DIMENSION D(1),E(1),Z(1)
      LOGICAL IQLW 
      DATA IQLW /.FALSE./
      DATA EPS/1.0D-16/,EPSS/1.0D-8/
      DATA DINF /1.0D20/, DINF2 /1.0D40/
      DATA DMIN /1.0D-20/
   
      IER=0
      IF (N .EQ. 1) RETURN
C *** 7.2.02 *** Uebernahme einer erweiterten Programmversion ***
      IF (IQLW .AND. IEV.EQ.1) THEN
         write(*,*) '*** IQLVMS-WARNUNG: Seit 7.2.02 neue Bedeutung',
     +	 ' des Parameters IEV ***'
         write(*,'(A,$)') '<ENTER> '
         read (*,*)
         IQLW=.FALSE.
      ENDIF	 
      ANORM=0.
      R = 0.0
C Alle Elemente der Nebendiagonale 1 nach vorn rcken
C ANORM := 1-Norm der Zeile mit grter 1-Norm
      DO 100 I = 2, N
         S = E(I)
         E(I-1) = S
         S = DABS(S)
         P = DABS(D(I-1)) + R + S
         IF (P .GT. ANORM)  ANORM = P
         R = S
  100 CONTINUE
      P = DABS(D(N)) + R
      IF (P .GT. ANORM)  ANORM = P
      E(N) = 0.0
      DO 250 L = 1, N
         J = 0
C     ********** LOOK FOR SMALL SUBDIAGONAL ELEMENT **********
  105    DO 110 M = L, N
            IF (M .EQ. N) GO TO 120
            IF (DABS (E(M)) .LE. EPS * (DABS (D(M)) + DABS (D(M+1)) ))
     1         GO TO 120
            IF (DABS (E(M)) .LE. DMIN * ANORM) GO TO 120
  110    CONTINUE
  120    P = D(L)
         MM1 = M - 1
         IF (M .EQ. L) GO TO 215
         IF (J .EQ. 30)  GO TO 1000
         J = J + 1
C     ********** FORM SHIFT **********
         G = (D(L+1) - P) / (2.0 * E(L))
         R = DSQRT (G * G + 1.0)
         S = P - E(L) / (G + DSIGN (R, G))
         IF (M .EQ. L+1)   GO TO 148
         T = S
         R = DMAX1 (DABS(S), ANORM/N)
         DO 140 I = 1, 6
            PA = P
            PSI = D(M) -T
            PSJ = - 1.0
            DO 130 KK = L, MM1
               K= L + MM1 -KK
               IF (DABS(PSI) .GE. EPS * DABS(E(K)))  GO TO 125
               PSI = DINF
               PSJ = DINF2
               GO TO 130
  125          P = E(K) / PSI
               PSI = D(K) - T - P * E(K)
               PSJ = P * P * PSJ - 1.0
  130       CONTINUE
            IF (DABS(PSJ) .LE. DMIN)  GO TO 148
            P = PSI / PSJ
            C = P
            IF (DABS(P) .GT. 0.5 * R)   C = DSIGN (R, P)
            T = T - C
            IF (DABS(P) .LE. EPSS * R)  GO TO 145
  140    CONTINUE
         GOTO  148
  145    S = T
  148    G = D(M) - S
         S = 1.0
         C = 1.0
         P = 0.0
         MML = M - L
         KI1 = MM1 * NM + 1
         KA = KI1 - NM
         KE = KA + IEV - 1
C     ********** FOR I=M-1 STEP -1 UNTIL L DO **********
         DO 200 II = 1, MML
            I = M - II
            F = S * E(I)
            B = C * E(I)
            IF (DABS(F) .LT. DABS(G)) GO TO 150
            C = G / F
            R = DSQRT (1D0 + C * C)
            E(I+1) = F * R
            S = 1.0 / R
            C = C * S
            GO TO 160
  150       S = F / G
            R = DSQRT (1D0 + S * S)
            E(I+1) = G * R
            C = 1.0 / R
            S = S * C
  160       G = D(I+1) - P
            R = (D(I) - G) * S + 2.0 * C * B
            P = S * R
            D(I+1) = G + P
            G = C * R - B
            IF (IEV .EQ. 0)  GOTO 200
C     ********** FORM VECTOR **********
            DO 180 KI=KA,KE
               F = Z(KI1)
               B = Z(KI)
               Z(KI1) = S * B + C * F
               Z(KI)  = C * B - S * F
               KI1 = KI1 + 1
  180       CONTINUE
            KI1 = KA
            KA  = KA - NM
            KE  = KE - NM
  200    CONTINUE
         D(L) = D(L) - P
         E(L) = G
         E(M) = 0.0
         GOTO 105
  215 CONTINUE
  250 CONTINUE
      RETURN
 1000 IER = L
      RETURN
      END
